// Code generated by "stringer -type=NEXTHOP_TYPE"; DO NOT EDIT.

package zebra

import "strconv"

const _NEXTHOP_TYPE_name = "FRR_NEXTHOP_TYPE_IFINDEXFRR_NEXTHOP_TYPE_IPV4FRR_NEXTHOP_TYPE_IPV4_IFINDEXFRR_NEXTHOP_TYPE_IPV6FRR_NEXTHOP_TYPE_IPV6_IFINDEXFRR_NEXTHOP_TYPE_BLACKHOLENEXTHOP_TYPE_IPV6_IFINDEXNEXTHOP_TYPE_IPV6_IFNAMENEXTHOP_TYPE_BLACKHOLE"

var _NEXTHOP_TYPE_index = [...]uint8{0, 24, 45, 74, 95, 124, 150, 175, 199, 221}

func (i NEXTHOP_TYPE) String() string {
	i -= 1
	if i >= NEXTHOP_TYPE(len(_NEXTHOP_TYPE_index)-1) {
		return "NEXTHOP_TYPE(" + strconv.FormatInt(int64(i+1), 10) + ")"
	}
	return _NEXTHOP_TYPE_name[_NEXTHOP_TYPE_index[i]:_NEXTHOP_TYPE_index[i+1]]
}
