// Code generated by "stringer -type=LSP_TYPE"; DO NOT EDIT.

package zebra

import "strconv"

const _LSP_TYPE_name = "LSP_NONELSP_STATICLSP_LDPLSP_BGPLSP_SRLSP_SHARP"

var _LSP_TYPE_index = [...]uint8{0, 8, 18, 25, 32, 38, 47}

func (i LSP_TYPE) String() string {
	if i >= LSP_TYPE(len(_LSP_TYPE_index)-1) {
		return "LSP_TYPE(" + strconv.FormatInt(int64(i), 10) + ")"
	}
	return _LSP_TYPE_name[_LSP_TYPE_index[i]:_LSP_TYPE_index[i+1]]
}
